''' Python wrapper class for representing and solving action graph games '''
import os, string, sys, time

AGG_SOLVER_PATH = "./bin/"
SOLVE_TIME = None
TIMERUN = False
TIME_LIMIT = 3600*2 # 2 hours
SOLVER = "simpdiv "
#SIMPDIV_OPTIONS = " -d -b -t -n0"
SIMPDIV_OPTIONS = ""
TEXT_FORMAT = True

def optimalVegas(total):
    ''' Generate an optimal sequence of runtimes < total '''
    seq = [1]
    new = 2
    sumSeq = 1
    while sumSeq < total:
        seq += seq + [new]
        sumSeq = sumSeq*2 + new
        new *=2
    while sumSeq > total:
        sumSeq -= seq[-1]
        del seq[-1]
    return seq

class AGG:
    def __init__(self,N,A,S,P,v,f,u,title=None):
        ''' N is a list of players
            A is a list of actions
            S is a mapping from players to lists of actions
            P is a list of projection (aka function) nodes
            v is a list of arcs (2-tupes of start,end nodes
            f is a mapping from projection node to type of projection (integer)
            u is a mapping from an action node to a payoff mapping (tuples of inputs to real values)
            '''
        self.N=N
        self.A=A
        self.S=S
        self.P=P
        self.v=v
        self.f=f
        self.u=u
        self.title = title
        self.testForExceptions()
        self.baseName = ".tmp"
        
    def testForExceptions(self):
        ''' Makes sure that the AGG is self-consistent '''
        return None # Can't run on Python 2.3 because there's no set object
        A2 = set()
        for actions in self.S.values():
            A2 = A2.union(actions)    
        for a in self.A:
            if a not in A2:
                raise "No agent can perform action "+str(a)

        for i in self.N:
            if not self.S.has_key(i):
                raise "No action set for player "+str(i)
        for i in self.S.keys():
            if i not in self.N:
                raise "Action set for non-existant player "+str(i)
            for a in self.S[i]:
                if a not in self.A:
                    raise "Action "+str(a)+" is not in A, but is in the action set of Player "+str(i)
        Nodes = self.A+self.P
        for start,end in self.v:
            if start not in Nodes:
                raise "Arc from non-existant node "+str(start)
            if end not in Nodes:
                raise "Arc to non-existant node "+str(end)

    def saveToFile(self, filename):
        if TEXT_FORMAT:
            self.saveToTextFile(filename)
        else:
            self.saveToBinaryFile(filename)
            
    def binWrite(self, file, variable, type='i'):
        import struct
        data = struct.pack(type, variable)
        file.write(data)

    def binWriteTuple(self, file, variable, type='i'):
        import struct
        data = struct.pack(type, *variable)
        file.write(data)
        
    ''' this should be storing the same info as the text version --- checked once'''
    def saveToBinaryFile(self,filename):
        Nodes = self.A+self.P
        file = open(filename,'wb')
        ''' use a 0 to indicate that this is a binary file and not text '''
        self.binWrite(file, 0, 'i')
        ''' length of title string followed by the string ( 0 ) if no title '''
        if self.title:
            self.binWrite(file, len(self.title))
            self.binWrite(file, self.title, str(len(self.title))+"s")
        else:
            self.binWrite(file, 0)
        ''' we have 3 numbers: n, A, P'''
        self.binWrite(file, len(self.N))
        self.binWrite(file, len(self.A))
        self.binWrite(file, len(self.P))
        ''' then we have n numbers specifying |A| for each player '''
        for i in self.N:
            self.binWrite(file, len(self.S[i]))
        ''' then we have the action indices for each player i'''
        for i in self.N:
            for s in self.S[i]:
                self.binWrite(file, self.A.index(s))
        ''' then we have, for each Node, the source of each arc '''
        for i in range(len(Nodes)):
            nodeName = str(Nodes[i])
            self.binWrite(file, len(nodeName))
            import struct
            data = struct.pack(str(len(nodeName))+"s", nodeName) #+1 for null
            file.write(data)
            arcs = filter(lambda arc: arc[1]==Nodes[i],self.v)
            self.binWrite(file,(len(arcs)))
            for start,end in arcs:
                self.binWrite(file,(Nodes.index(start)))
        ''' then we have P numbers, each representing the type of each function node '''
        for n in self.P:
            self.binWrite(file,(self.f[n]))
        ''' now for each action node, have
        name (null terminated string),
        the number 1
        some utility value of some sort
        string that represents the key 
        utility
        '''
        import locale
        for a in self.A: #a is a string
            #file.write(str(a))
            self.binWrite(file,1)
            self.binWrite(file,(len(self.u[a].keys())))
            keys = self.u[a].keys()
            keys.sort() # For neatness
            for key in keys:
                self.binWrite(file,len(key)); # how many elements in key?
                for ele in key:
                    if type(ele) == 'str':
                        numeric = string.atoi(ele)
                    else:
                        numeric = int(ele)
                    self.binWrite(file,numeric)
                self.binWrite(file, self.u[a][key], 'd')
        file.close()
        
        
    def saveToTextFile(self,filename):
        ''' Saves the game to a formal readible by Albert Xin Jiang's solvers '''
        Nodes = self.A+self.P
        file = open(filename,'w')
        ''' the first thing we have is a null-terminated string that is the title'''
        if self.title:
            file.write("# %s\n\n"%(self.title))
        ''' then we have 3 numbers: n, A, P'''
        file.write("# Players\n%d\n# Action Nodes\n%d\n# Function nodes\n%d\n"%(len(self.N),len(self.A),len(self.P)))
        ''' then we have n numbers specifying |A| for each player '''
        file.write("# |Action sets|\n")
        for i in self.N:
            file.write(str(len(self.S[i]))+" ")
        ''' then we have the action indices for each player i'''
        file.write("\n\n# Action sets\n")
        for i in self.N:
            for s in self.S[i]:
                file.write(str(self.A.index(s))+" ")
            file.write("\n")
        ''' then we have, for each Node, the source of each arc '''
        file.write("\n\n# Arcs\n")
        for i in range(len(Nodes)):
            n = Nodes[i]
            file.write("# Node %d: %s\n"%(i,str(n)))
            arcs = filter(lambda arc: arc[1]==n,self.v)
            file.write(str(len(arcs))+" ")
            for start,end in arcs:
                file.write(str(Nodes.index(start))+" ")
            file.write("\n")
        ''' then we have P numbers, each representing the type of each function node '''
        file.write("\n\n# Types of function nodes\n")
        for n in self.P:
            file.write(str(self.f[n])+" ")
        ''' now for each action node, have
        name (null terminated string),
        the number 1
        some utility value of some sort
        string that represents the key 
        utility
        '''
        file.write("\n\n# Payoffs of action nodes\n")
        for a in self.A: #a is a string
            file.write("# Action node: %s\n"%(str(a)))
            file.write("1 "+ str(len(self.u[a].keys()))+"\n") # how many keys
            keys = self.u[a].keys()
            keys.sort() # For neatness
            for key in keys:
                try:
                    keyStr = string.join(map(str,key))
                except Exception:
                    keyStr = str(key)
                file.write("  [ "+keyStr+" ] "+str(self.u[a][key])+"\n")
        file.close()
        
        

    def generateStart(agg,filename,seed):
        output = ""
        import random
        random.seed(seed)
        for i in agg.N:
            s = len(agg.S[i])
            j = random.choice(range(s))
            strategy = "0,"*j+"1,"+("0,"*(s-1-j))
            output += strategy
        open(filename,'w').write(output[:-1]+"\n")       

    def iterativeSolve(self,startSeed=1,scale=1):
        ''' Returns a Nash equilibrium (a mapping from players to strategies where a strategy is a mapping from action to probability) '''
        aggTempFile = self.baseName+".agg"
        neTempFile = self.baseName+".ne"
        timeTempFile = self.baseName+".time"
        startTempFile = self.baseName+".start"
        self.saveToFile(aggTempFile)
        global SOLVE_TIME
        SOLVE_TIME = 0.0
        for limit in optimalVegas(TIME_LIMIT/scale):
            #print "\t",limit
            self.generateStart(startTempFile,startSeed)
            startSeed+=1
            os.system(AGG_SOLVER_PATH+"timerun "+str(limit*scale)+" "+AGG_SOLVER_PATH+"simpdiv "+SIMPDIV_OPTIONS+" <"+aggTempFile+" 1>"+neTempFile+" 2>"+timeTempFile)
            timedata = open(timeTempFile,'r').read()
            SOLVE_TIME += float( string.split(timedata,"TIME USED: ")[1] )
            if string.find(timedata,"c child exit by a signal") == -1:
                text = open(neTempFile,'r').read()
                if text[:3] != "NE,":
                    raise "Invalid output from simpdiv!: "+text    
                return self.__decodeStrategies(text[3:])
        raise "Timeout"
        
    def multisolve(self,n,seedStart=0):
        def myspawn(commandline):
            pid = os.fork()
            if pid == 0:
                sys.exit(os.system(commandline))
            else:
                return pid
        ''' Returns a Nash equilibrium (a mapping from players to strategies where a strategy is a mapping from action to probability) '''
        aggTempFile = self.baseName+".agg"
        self.saveToFile(aggTempFile)
        global SOLVE_TIME
        pids = [0]*n
        SOLVE_TIME = 0.0
        limit = TIME_LIMIT
        while True:
            for i in range(n):
                timeTempFile = self.baseName+".time"+str(i)
                neTempFile = self.baseName+".ne"+str(i)
                startTempFile = self.baseName+".start"+str(i)
                self.generateStart(startTempFile,seedStart)
                seedStart += 1
                cmd = AGG_SOLVER_PATH+"timerun "+str(limit)+" "+AGG_SOLVER_PATH+"simpdiv -s "+startTempFile+" <"+aggTempFile+" 1>"+neTempFile+" 2>"+timeTempFile
                #print cmd
                pids[i] = myspawn(cmd)
            for i in range(n):
                pid,code = os.wait()
                print pid
                i = pids.index(pid)
                timeTempFile = self.baseName+".time"+str(i)
                neTempFile = self.baseName+".ne"+str(i)
                try:
                    text = open(neTempFile,'r').read()
                except: continue
                if text[:3] != "NE,":
                    continue
                SOLVE_TIME += float( string.split(open(timeTempFile,'r').read(),"TIME USED: ")[1] )*n
                for pid in pids:
                    os.kill(pid,9)
                return self.__decodeStrategies(text[3:])
            SOLVE_TIME += n*limit
            limit *= 2
            raise "Couldn't do it in n attempts!" # FIXME: Comment this out when I want it to keep going

    def solve(self):
        ''' Returns a Nash equilibrium (a mapping from players to strategies where a strategy is a mapping from action to probability) '''
        # FIXME: Let the user specify the solver
        aggTempFile = self.baseName+".agg"
        neTempFile = self.baseName+".ne"
        self.saveToFile(aggTempFile)
        global SOLVE_TIME
        if TIMERUN: #
            timeTempFile = self.baseName+".time"
            os.system(AGG_SOLVER_PATH+"timerun "+str(TIME_LIMIT)+" "+AGG_SOLVER_PATH+"simpdiv "+SIMPDIV_OPTIONS+" <"+aggTempFile+" 1>"+neTempFile+" 2>"+timeTempFile)
            SOLVE_TIME = float( string.split(open(timeTempFile,'r').read(),"TIME USED: ")[1] )
        else:
            startTime = time.time()
            os.system(AGG_SOLVER_PATH+SOLVER+SIMPDIV_OPTIONS+" <"+aggTempFile + " >"+neTempFile);
            SOLVE_TIME = time.time()-startTime
            print "solve time: " + str(SOLVE_TIME)
        text = open(neTempFile,'r').read()
        
        NEs = []
        lines = text.splitlines()
        for line in lines:
            if line[:3] != "NE,":
                print "Warning, unexpected line in NE file " + line 
#                raise "Invalid output from " + SOLVER + "!: " + line
            else:
                NEs.append( self.__decodeStrategies(line[3:]) )
        return NEs
    
    def payoffs(self,strategyProfile):
        ''' Returns the expected payoffs of each player given the strategy profile '''
        stratStr = self.__encodeStrategies(strategyProfile)
        open(self.baseName+".ne","w").write(stratStr+"\n")
        self.saveToFile(self.baseName+".agg")
        os.system(AGG_SOLVER_PATH+"getpayoffs "+self.baseName+".agg <"+self.baseName+".ne >"+self.baseName+".po")
        text = open(self.baseName+".po","r").readlines()
        return self.__decodePayoffs(text[1])

    def socialWelfare(self,strategyProfile):
        ''' Returns the expected social welfare given the strategy profile '''
        payoffs = self.payoffs(strategyProfile)
        return sum([(payoffs[i]) for i in payoffs.keys()])
            


    # Just some utilities for interacting with external programs
    def __encodeStrategies(self,strategyProfile):
        out = ""
        for i in self.N:
            for a in self.S[i]:
                if strategyProfile[i].has_key(a):
                    p = strategyProfile[i][a]
                else: p = 0.0
                out += "%.5f "%(p)
        return out
                    
    def __decodeStrategies(self,solnStr,allActions=False,delimiter=","):
        parts = string.split(solnStr,delimiter)
        j = 0
        eq = {}
        for i in self.N:
            eq[i] = {}
            for s in self.S[i]:
                try:
                    value = float(parts[j])
                except Exception:
                    top,bottom = string.split(parts[j],"/")
                    value = float(top)/float(bottom)
                if value > 0.0 or allActions:
                    eq[i][s] = value
                j+=1
        return eq

    def __decodePayoffs(self,payoffStr,delimiter=" "):
        ''' Takes a delimited string of payoffs and converts it into a dictionary '''
        parts = string.split(payoffStr,delimiter)
        payoffs = dict([
            (self.N[i], float(parts[i]))
                for i in range(len(self.N))])
        return payoffs

    
    def payoffs(self,strategyProfile):
        ''' Returns the expected payoffs of each player given the strategy profile '''
        stratStr = self.__encodeStrategies(strategyProfile)       
        return self.__decodePayoffs(PayoffComputer(self.baseName+".agg").test(stratStr))        

    def isNE(self,strategyProfile):
        stratStr = self.__encodeStrategies(strategyProfile)
        open(self.baseName+".ne","w").write(stratStr+"\n")
        self.saveToFile(self.baseName+".agg")
        aSizes = [(len(self.S[i])) for i in self.N]
        return PayoffComputer(self.baseName+".agg",aSizes).isNE(stratStr)
   
    
import re
def Find_aSizes(aggFile):
    ''' Returns the size of the action set of each player '''
    data = open(aggFile,'r').readlines()
    data = map(lambda x: string.strip(re.sub("#.*","",x)), data)
    data = filter(lambda x: len(x)>0,data)
    data = string.split(string.join(data))
    n = int(data[0])
    aSizes = data[3:n+3]
    return map(int,aSizes)

class PayoffComputer:
   def __init__(self,aggFilename,aSizes=None):
       (self.tochild,self.fromchild) = os.popen2(AGG_SOLVER_PATH+"getpayoffs "+aggFilename)
       if aSizes == None: aSizes = Find_aSizes(aggFilename)
       self.aSizes = aSizes

   def fixStrategy(self,stratStr):
       stratStr = string.strip(stratStr)
       if stratStr[:3] == "NE,": stratStr = stratStr[3:]
       strat = string.split(stratStr,",")
       if len(strat)==1:
           strat = string.split(stratStr)
       strat2 = []
       for item in strat:
           try:
               top,bottom = string.split(item,"/")
               strat2.append(str(float(top)/float(bottom)))
           except:
               strat2.append(item)
       return string.join(strat2)

   def test(self, strategyString):
       strategyString = self.fixStrategy(strategyString)
       self.tochild.write(strategyString+"\n")
       self.tochild.flush()
       self.fromchild.readline()
       return string.strip(self.fromchild.readline())

   def parse(self,strategyString):
       strategyString = self.fixStrategy(strategyString)
       output = []
       nPreviousActions=0
       for i in range(len(self.aSizes)):
           nA = self.aSizes[i]
           strategy = string.split(strategyString)[nPreviousActions:nPreviousActions+nA]
           output.append(map(float,strategy))
           nPreviousActions += nA
       return output

   def isNE(self, strategyString):
       # FIXME: Change this to use parse()
       nPreviousActions = 0
       eu = map(float,string.split(self.test(strategyString)))
       for i in range(len(self.aSizes)):
           nA = self.aSizes[i]
           prefix = string.split(strategyString)[0:nPreviousActions]
           suffix = string.split(strategyString)[nPreviousActions+nA:]
           for a in range(nA):
               strategy = ["0"]*a+["1"]+["0"]*(nA-a-1)
               newStrategy = string.join(prefix+strategy+suffix)
               payoff = float(string.split(self.test(newStrategy))[i])
               if payoff > eu[i]:
                   return False
           nPreviousActions += nA
       return True

   def kill(self):
       self.tochild.close()
       self.fromchild.close()
def pcTest():
    pc = PayoffComputer("PD.agg")
    print pc.test("1 0 1 0")
    
def tempFileName():
    import random
    return str(random.random())

def binRead(file, type='i'):
    import struct
    data = file.read(struct.calcsize(type))
    return struct.unpack(type, data)

''' implement createFromTextFile() if needed '''
#def createFromFile(filename):
#    return createFromBinFile(filename)

#def createFromBinFile(filename):
#    file = open(filename, 'rb')
#    ''' first 4 bytes indicates if this is binary file '''
#    import struct
#    file.seek(struct.calcsize('i'))
#    ''' there may or may not be a title '''
#    titleLen = binRead(file)[0]
#    title = None
#    if titleLen > 0:
#         title = binRead(file, str(titleLen)+"s")[0]
#    ''' we have 3 numbers: n, A, P'''
#    n = binRead( file )[0]
#    # NOTE THAT WHILE AGG STORES PLAYER 'NAMES', THEY ARE NOT STORED IN THE AGG FILE
#    N = range(n)
#    a = binRead( file )[0]
#    A = []
#    p = binRead(file)[0]
#    P = []
#    ''' then we have n numbers specifying |A| for each player '''
#    NumActions = []
#    S = {}
#    for i in N:
#        S[i] = []
#        numActionsForI = binRead(file)
#        NumActions.extend(numActionsForI)
#    ''' then we have the action indices for each player i'''
#    ActionIndex = []
#    for i in N:
#        ActionIndex.append([])
#        for s in range(NumActions[i]): #for each action of the player 
#            ActionIndex[i].append( binRead(file)[0] )
#            # moved this to after in order to get the name S[i].append(actionIndex)
#    ''' then we have, for each Node, the source of each arc '''
#    v = []
#    NumArcs = []
#    NodeNames = []
#    
#    for i in range(a+p):
#        '''store the node's name'''
#        nameLen = binRead(file)[0]
#        NodeNames.append(binRead(file, str(nameLen)+"s" )[0])
#        NumArcs.extend( binRead(file))
#        for arcIndex in range(NumArcs[i]):
#            arcStart = binRead(file)[0]
#            v.append((arcStart,i))
#            
#    ''' put the names in the arcs instead of indices '''
#    for i in range(len(v)): 
#        v[i] = (NodeNames[v[i][0]], NodeNames[v[i][1]])
#                
#    ''' then we have P numbers, each representing the type of each function node '''
#    f = {}
#    for fn in range(p):
#        P.append(NodeNames[a+fn])
#        functionType = binRead(file)[0]
#        f[P[fn]] = functionType
#            
#    ''' now for each action node, have
#    name (null terminated string),
#    the type
#    some utility value of some sort
#    string that represents the key 
#    utility
#    '''
#    u = {}
#    for aIndex in range(a):
#        a = NodeNames[aIndex]
#        A.append(a)
#        u[a] = {}
#        type = binRead(file)[0]
#        numKeys = binRead(file)[0]
#        
#        for keyIdx in range( numKeys ):
#            ''' need to know length of key'''
#            keylen = binRead(file, 'i')[0]
#            key = []
#            for ele in range (keylen):
#                keyele = binRead(file, 'i')[0]
#                key.append(keyele)
#            value = binRead(file, 'd')[0]
#            u[a][tuple(key)] = value
#   
#    file.close()
#    
#    for i in N:
#        for s in range(NumActions[i]): #for each action of the player 
#            S[i].append(A[ActionIndex[i][s]])
#            
#    return AGG(N,A,S,P,v,f,u,title)

def PrisonersDilemma():
    # Simple test: Prisoner's dilemma
    N_ = [1,2]
    A_ = ["C","D"]
    S_={1:"CD",2:"CD"}
    P_=[]
    v_ = ["CC","DD"]
    f_={}
    u_ = {
            "C":{"1":0,"2":3},
            "D":{"1":5,"2":1}
#            "C":{"1":1,"2":2},
#            "D":{"1":1,"2":2}
        }
    agg = AGG(N_,A_,S_,P_,v_,f_,u_,"Prisoner's dilemma")
    return agg

def RockPaperScissors():
    N_ = [1,2,3]
    A_ = ["R","P","S"]
    S_={1:"RPS",2:"RPS",3:"RPS"}
    P_=[]
    v_ = ["PR","SP","RS","RP","PS","SR"]
    f_={}
    u_ = {
            "R":{"00":0,"01":1,"10":-1,"11":0,"02":2,"20":-2},
            "P":{"00":0,"01":1,"10":-1,"11":0,"02":2,"20":-2},
            "S":{"00":0,"01":1,"10":-1,"11":0,"02":2,"20":-2},
        }
    agg = AGG(N_,A_,S_,P_,v_,f_,u_,"Rock Paper Scissors")
    return agg

if __name__=="__main__":
#    agg = PrisonersDilemma()
#    
#    agg.saveToFile("PD.agg")
#    agg.saveToTextFile("PD.agg.txt")
#    agg2 = createFromFile("PD.agg")
#    agg2.saveToTextFile("PD2.agg.txt")
    
    agg = RockPaperScissors()
#    agg.saveToFile("RPS.agg")
#    agg.saveToTextFile("RPS.agg.txt")
#    agg2 = createFromFile("RPS.agg")
#    agg2.saveToTextFile("RPS2.agg.txt")
    NEs= agg.solve()
    for ne in NEs:
        print "Nash equilibrium of game: ",ne
        print "Payoffs: ",agg.payoffs(ne)
    print "Compute time (s): ",SOLVE_TIME
    
